import base64
from Crypto.Cipher import AES
from urlparse import urlparse
import ast
from elasticsearch import helpers
from elasticsearch import Elasticsearch
import json
import os
import sys
import yaml
from scriptConst import Constants
import requests
import json



def iterateDict(data):
    local_dict = {}
    if type(data) != list:
        for key,val in data.items():
            if type(val) == unicode:
                local_dict[key.encode("utf-8")] = val.encode("utf-8")
            elif type(val) != dict:
                local_dict.update({key.encode("utf-8"):val})
            elif type(val) == dict:
                local_dict.update({key.encode("utf-8"):iterateDict(val)})
            else:
                local_dict.update({key.encode("utf-8"):val})
    else:
        li = []
        for i in data:
            li.append(iterateDict(i))
        return li

    return local_dict




def decrypt(enc):
    def unpad(raw_data):
        return raw_data[:-ord(raw_data[len(raw_data)-1:])]
    enc = base64.b64decode(enc)
    iv = enc[:16]
    cipher = AES.new(Constants.decrypt_key, AES.MODE_CBC, iv)
    return unpad(cipher.decrypt( enc[16:] ))

def prepare_data( target, data):
    DOCUMENTTYPE = Constants._DOC
    docs = []
    config = target.get(Constants.CONFIG, {})
    if config and config.get(Constants.ENABLED):
        index = target[Constants.CONFIG][Constants.INDEX]
        for document in data:
            doc = {}
            doc[Constants._INDEX] = index + Constants._WRITE
            doc[Constants._TYPE] = DOCUMENTTYPE
            doc[Constants._SOURCE] = document
            docs.append(doc)
        return docs
    else:
        # logger.error("Target is not enabled "+target.get(Constants.NAME, ""))
        print("error while forming data")

def get_es_client( target):
    config = target.get(Constants.CONFIG, {})
    if config:
        http_auth = ''
        host = config.get(Constants.HOST)
        port = config.get(Constants.PORT)
        #host = '34.219.3.98'
        if host:
            scheme = config.get(Constants.PROTOCOL)
            username = config.get(Constants.USERNAME)
            password = config.get(Constants.PASSWORD)
            if password:
                password = base64.b64decode(password)
            if username and password:
                http_auth = (username,password)
            es = Elasticsearch(hosts=[{Constants.HOST: host, Constants.PORT: port}],
                                http_auth=http_auth,
                                scheme=scheme
                                )
            return es
        else:
            # logger.error("Target Dosen't contain host " +
                            # target.get(Constants.NAME, ""))
            print("tagret doesnt contain host")
            return
    else:
        # logger.error("Target Dosen't contain config " +
                        # target.get(Constants.NAME, ""))
        print("target doesnt contain config")
        return

def write_docs_bulk(config,data):


    res1 = get_index_conf(decrypt(config["key"]),config["tags"]["projectName"],'metric' )
    targets = [iterateDict(res1)]


    MAXRETRY = 2
    TIMEOUT = 10
    # response_output = []
    for target in targets:
        

        
        # target["config"]["password"] = "2330996677315"
        # target["config"]["username"] = "apmuser"
        cfg = target.get(Constants.CONFIG ,{})
        if cfg and cfg.get(Constants.TYPE) == Constants.ELASTICSEARCH:
             
            # target["config"]["host"] = "34.214.173.28"
            es_client = get_es_client(target)
            docs1 = prepare_data(target, data)
            index_search = docs1[0]["_index"].split("$")[0] + "*"
      
            if not es_client.ping():
                pass

            docs = prepare_data(target, data)
            if es_client and docs:
                try:
                    res = helpers.bulk(es_client, docs, refresh=True, request_timeout=TIMEOUT, max_retries=MAXRETRY)
                    return res
                except Exception as e:
                    print("Issue with indexing data",e)
        else:
            send_docs_to_kafka_rest_proxy(target,data)
    return


def send_docs_to_kafka_rest_proxy(target,data):
    cfg = target.get(Constants.CONFIG ,{})
    print(data)
    if cfg and data:
        url = cfg.get("url","")
        index = cfg.get(Constants.INDEX)

        url = url + "/topics/" + index
        print(url)
        if url:
            username = cfg.get(Constants.USERNAME)
            password = cfg.get(Constants.PASSWORD)
            auth_token = cfg.get(Constants.TOKEN)
            if password:
                password = base64.b64decode(password)
            if auth_token:
                headers = {"Content-Type":"application/vnd.kafka.json.v2+json","Accept":"application/vnd.kafka.v2+json","Authorization":auth_token}
            else:
                headers = {"Content-Type":"application/vnd.kafka.json.v2+json","Accept":"application/vnd.kafka.v2+json"}
            auth=(username,password)
            #preparing payload for kafka
            payload = {}
            payload["records"] = []
            for document in data:
                value = {}
                value["value"]=document
                payload["records"].append(value)
            try:
                r = requests.post(url, data=json.dumps(payload), headers=headers,timeout=5,verify=False)
                print("status", r.status_code)
                if r.status_code !=200:
                    print("Status Code in kafka code: %s" , str(r.status_code))
                    # logger.error("Status Code: %s" , str(r.status_code))
                    # logger.error(r.text)
                
            except Exception as e:
                print("Exception kafka post: %s" , e)
                # logger.error(e)
                # logger.error("Issue with sending data to kafka-rest")
    return

def get_index_conf(conf,project_name = "", plugin_type = ""):

    es_conf = {}
    conf  = json.loads(conf)
    tar_type = conf.get(Constants.TYPE)
    profile_id = conf[Constants.PROFILE_ID]
    owner = conf.get(Constants.OWNER,"")


    if owner:
        # self.owner = owner
        del conf[Constants.OWNER]
    del conf[Constants.PROFILE_ID]
    # name = Constants.CONTROL + "-" + profile_id
    name = plugin_type + "-" + profile_id
    if conf.get(Constants.PASSWORD):
        conf[Constants.PASSWORD] = base64.b64encode(str(conf[Constants.PASSWORD]))

    if tar_type == Constants.ELASTICSEARCH:
        es_conf[Constants.TYPE] = Constants.ELASTICSEARCH
        if plugin_type:
            name = plugin_type + "-" + profile_id + "-" + project_name

    es_conf[Constants.NAME] = name
    conf[Constants.INDEX] = name
    conf[Constants.ENABLED] = True
    es_conf[Constants.CONFIG] = conf

    return es_conf




import schedule
import time
import ast
import os
import json
import requests
from pushtoES import write_docs_bulk

from datetime import datetime
import datetime as dt
import calendar
from datetime import datetime, timedelta
from dateutil.tz import tzutc, tzlocal
import random
import dateutil.relativedelta
import threading

def iterateDict(data):
    local_dict = {}
    for key,val in data.items():
        if type(val) == unicode:
            local_dict[key.encode("utf-8")] = val.encode("utf-8")
        elif type(val) != dict:
            local_dict.update({key.encode("utf-8"):val})
        elif type(val) == dict:
            local_dict.update({key.encode("utf-8"):iterateDict(val)})
        else:
            local_dict.update({key.encode("utf-8"):val})
    return local_dict

def get_data(urlDict):
    tasksUrl = urlDict.get("url")
    all_data = []
    for url in [tasksUrl]:
        response = requests.get(
            url,
            # params={'duration': '10d'},
            headers={'Authorization': urlDict["authkey"],'Content-type': 'application/json' },
            verify=False
        )
        data = json.loads(response.text)
        # data = ast.literal_eval(json.dumps(data))

        all_data = all_data + data.get("result", [])
    return all_data

def form_task_record(data, flag, tags, curr_time, taskId, bufferData=None):
    metadata = {
         "SCRIPT_NAME": data["SCRIPT_NAME"],
         "TABLE_NAME": data["TABLE_NAME"],
         "LOG_DESC": data["LOG_DESC"],
         "ADDITIONAL_INFO": data["ADDITIONAL_INFO"],
         "IP_ADDRESS": data["IP_ADDRESS"],
         "MODULE_CODE":data["MODULE_CODE"],
         "WORKSTATION": data["WORKSTATION"],
         "HOST_NAME": data["HOST_NAME"]
    }
    rec = {
        "jobName" : data.get("ETL_EXECUTION_ID"),
        "jobId" : data["jobId"],
        "stageName":data["stageName"],
        "time": curr_time,
        "taskName": taskId,
        "status": data['STATUS'],
        "_documentType":"task"
    }
    rec['metaData'] = metadata
    rec.update(tags)
    if flag == 'started':
        rec['startTime'] = data['time']
        rec['started'] = True
    if flag == 'completed':
        if taskId in bufferData.get(data['jobId']).keys():
            record = bufferData.get(data['jobId']).get(taskId)
            rec['duration'] = data['time'] - record['startTime']
            rec['endTime'] = data['time']
            rec['stageName'] = record['stageName']
            rec['startTime'] = record['startTime']
            rec['completed'] = True
        else:
            print("no start record found")
            rec = None
    return rec

def form_job_record(data, flag, tags, curr_time, bufferData=None):
    metadata = {
         "SCRIPT_NAME": data["SCRIPT_NAME"],
         "TABLE_NAME": data["TABLE_NAME"],
         "LOG_DESC": data["LOG_DESC"],
         "ADDITIONAL_INFO": data["ADDITIONAL_INFO"],
         "IP_ADDRESS": data["IP_ADDRESS"],
         "MODULE_CODE":data["MODULE_CODE"],
         "WORKSTATION": data["WORKSTATION"],
         "HOST_NAME": data["HOST_NAME"]
    }
    rec = {
        "jobName" : data["ETL_EXECUTION_ID"],
        "jobId" : data["jobId"],
        "startTime": data["time"],
        "time": curr_time,
        "status": data['STATUS'],
        "_documentType":"job",
    }
    rec['metaData'] = metadata
    rec.update(tags)
    if flag == 'started':
        rec['started'] = True
    else:
        rec['completed'] = True
        rec['endTime'] = data['time']
        if 'job' in bufferData.get(data['jobId']).keys():
            record = bufferData.get(data['jobId'])['job']
            rec['duration'] = data['time'] - record['startTime']
            rec['startTime'] = record['startTime']
        else:
            print("No job start record found")
    return rec

def form_stage_record(data, flag, tags):
    rec = {
        "jobName" : data["ETL_EXECUTION_ID"],
        "jobId" : data["jobId"],
        "startTime": data["time"],
        "time": int(time.time() * 1000),
        "status": data['STATUS'],
        "_documentType":"stage",
        "stageName": data['stageName']
    }
    rec.update(tags)
    if flag == 'started':
        rec['started'] = True
    else:
        record = bufferData.get(data['jobId']).get(data['stageName'])
        rec['duration'] = data['time'] - record['startTime']
        rec['endTime'] = data['time']
        rec['startTime'] = record['startTime']
        rec['completed'] = True
    return rec


def work(metricConfig, bufferPath):

    tags = {
        "_plugin": (metricConfig.get("metrics",{}).get("plugins")[0]).get("name"),
        "_tag_Name": metricConfig.get("tags").get("Name"),
        "_tag_appName": metricConfig.get("tags").get("appName"),
        "_tag_projectName": metricConfig.get("tags").get("projectName")
    }

    datalogs = get_data(metricConfig.get("metrics",{}).get("plugins")[0])
  
    # datalogs = ast.literal_eval(json.dumps(datalogs))
    datalogs = [iterateDict(i) for i in datalogs]

    # datalogs = [i for i in datalogs if i.get("job_id",0) == "854e0a8d-3125-44c2-935b-0e219ec262d992"]
    curr_time = int(time.time() * 1000) 
    
  
    def sort_by_time(data):
        return data['time']

    datalogs.sort(key=sort_by_time, reverse=False)
   
    started_rec = [x for x in datalogs if 'start' in x['STATUS'].lower()]
    completed_rec = [x for x in datalogs if 'start' not in x['STATUS'].lower()]
    

    started_records = {}
    completed_records = {}
    task_exclusion_list = ["1828591 MAIN INCR ETL", "1828594 E2B INCR ETL", "1828595 SET_START_PARAM", "1828599 SET_START_PARAM", "1830330 ETL_HOOK_POST_CDR_TRANSFORM-INCR", "TEMP_REPORTABILITY_FLAG", "1829247 ACTIVESUBSTANCE_ADDL", "1829244 PRIMARYSOURCE_ADDL", "1829241 LITERATUREREFERENCES_ADDL", "1829238 RECEIVER_ADDL", "1829235 MHLWADMICSRCASENUM_ADDL", "1829233 DRUGREACTIONRELATEDNESS_ADDL", "1829230 MHLWDUMMY_ADDL", "1829228 DRUG_ADDL", "1829225 DOSAGEINFORMATION_ADDL", "1829222 PATIENT_ADDL", "1829219 REACTION_ADDL", "1829979 T_IDENTIFICATION_FU", "1829975 T_IDENTIFICATION_FU", "1829967 S_IDENTIFICATION_FU", "1830287 CDR_PROD_AE_TTO"]
    completedJob = []
    with open(bufferPath, 'r') as f:
       bufferData = json.load(f)
    for data in started_rec:
        if data['jobId'] not in bufferData.keys():
            bufferData[data['jobId']] = {}
        jobId = data['jobId']
        if 'job' in bufferData.get(data['jobId']).keys():
            print("i1")
            record = bufferData.get(data['jobId'])['job']
            duration = data['time'] - record['startTime']
            print(duration)
        if data.get("TABLE_NAME","") == "INCR ETL":
            if 'job' not in bufferData[jobId].keys():
                record = form_job_record(data, 'started', tags, curr_time)
                started_records.update({jobId:record})
                bufferData[jobId].update({'job': record})
        else:
            if data.get("ID") is not None:
                taskId = str(int(data["ID"])) + ' ' + data["TABLE_NAME"]
            else:
                taskId = data["TABLE_NAME"]
            if taskId not in task_exclusion_list:
                record = form_task_record(data, 'started', tags, curr_time, taskId)
                if taskId not in bufferData.get(jobId).keys():
                    bufferData[jobId].update({taskId: record})
                    id_1 = data['jobId'] + '_' + taskId
                    started_records.update({id_1:record})
    for data in completed_rec:
        jobId = data['jobId']
        if data.get("TABLE_NAME","") == "INCR ETL":
            if data['jobId'] in started_records.keys():
                completed_record = started_records[jobId]
                completed_record['duration'] = data['time'] - completed_record['startTime']
                completed_record['time'] = curr_time
                completed_record['endTime'] = data['time']
                del completed_record['started']
                completed_record['status'] = data['STATUS']
                completed_record['completed'] = True        
                del started_records[jobId]
                completedJob.append(jobId)
                completed_records.update({jobId: completed_record}) 

            else:    
                record = form_job_record(data, 'completed', tags, curr_time, bufferData)
                completed_records.update({data['jobId']:record})
                completedJob.append(jobId)
            
        else:
            if data.get("ID") is not None:
                taskId = str(int(data["ID"])) + ' ' + data["TABLE_NAME"]
            else:
                taskId = data["TABLE_NAME"]
            id_1 = data['jobId'] + '_' + taskId
            if taskId not in task_exclusion_list:
                """
                Common records - directly insert completed records
                """
                new_task = data['jobId'] + '_' + taskId
                if new_task in started_records.keys():
                    if data['jobId'] == started_records.get(id_1)['jobId']:
                        completed_record = started_records[id_1]
                        completed_record['duration'] = data['time'] - completed_record['startTime']
                        completed_record['time'] = curr_time
                        completed_record['endTime'] = data['time']
                        del completed_record['started']
                        completed_record['status'] = data['STATUS']
                        completed_record['completed'] = True        
                        del started_records[id_1]
                        del bufferData[data['jobId']][taskId]
                        completed_records.update({id_1: completed_record}) 
                else:
                    rec = form_task_record(data, 'completed', tags, curr_time, taskId, bufferData)
                    if rec is not None:
                        del bufferData[data['jobId']][taskId]
                        completed_records.update({id_1: rec})

    for jobId in completedJob:
        taskKeys = bufferData.get(jobId).keys()
        for taskId in taskKeys:
            if taskId != 'job':
                id_1 = jobId + '_' + taskId
                rec = bufferData.get(jobId).get(taskId)
                rec['endTime'] = None
                rec['completed'] = True
                rec['duration'] = None
                rec['status'] = None
                completedJob = completed_records.get(jobId)
                rec['time'] = curr_time
                del rec['started']
                completed_records.update({id_1: rec})
        del bufferData[jobId]
    with open(bufferPath, "w") as fl:
        json.dump(bufferData,fl)
    for _, val in completed_records.items():
        write_docs_bulk(metricConfig, [val])
    
    for _, val in started_records.items():
        write_docs_bulk(metricConfig, [val])



class Constants:

  _DOC =  "_doc"
  _INDEX =  "_index"
  _SOURCE =  "_source"
  _TYPE =  "_type"
  _WRITE =  "-$_write"
  CONFIG =  "config"
  CONTROL =  "control"
  ELASTICSEARCH =  "elasticsearch"
  ENABLED =  "enabled"
  HOST =  "host"
  INDEX =  "index"
  NAME =  "name"
  OWNER =  "owner"
  PASSWORD =  "password"
  PORT =  "port"
  PROFILE_ID =  "profile_id"
  PROTOCOL =  "protocol"
  TYPE =  "type"
  USERNAME =  "username"
  decrypt_key =  "SnappyFlow123456"
  TOKEN = 'token'




import click
import os
import yaml
import time
import schedule
import sys
import json
import inp as inp

@click.command()
@click.argument("config_path", default='', required=False)
def main(config_path):

    config = {}
    try:
        print(config_path)
        # if os.path.isfile("config.yaml"):
        with open(config_path) as file:
            config = yaml.load(file)
                # logger.info(config)
        bufferPath = "bufferdata.json"
        if "/" in config_path:
            bufferPath = "/".join(config_path.split("/")[:-1]) + "/bufferdata.json"
        if os.path.exists(bufferPath):
            with open(bufferPath, "r") as f:
                try:
                    prev_data = json.load(f)
                except ValueError:
                    prev_data = {}

            with open(bufferPath, "w") as fl:
                # json.dumps(datalogs , fl)
                json.dump(prev_data, fl)
        else:

            with open(bufferPath, "w") as fl:
                json.dump({},fl)

        inp.work(config, bufferPath)

    except Exception as exception:

        print("error in opening config.yaml", str(exception))




"""
etlReport script
"""
from setuptools import find_packages, setup
import io
import os
import re

def read(filename):
    filename = os.path.join(os.path.dirname(__file__), filename)
    text_type = type(u"")
    with io.open(filename, mode="r", encoding='utf-8') as fd:
        return re.sub(text_type(r':[a-z]+:`~?(.*?)`'), text_type(r'``\1``'), fd.read())

dependencies = ['click','elasticsearch','schedule','requests','pycrypto',"PyYAML","urllib3"]

setup(
    name='sfapmetl',
    version='0.2.5',
    url='https://github.com/ruchira-dot-rao/sf-apm-etl',
    license='BSD',
    author='dummy',
    author_email='ruchira.rao@maplelabs.com',
    description='etlReport script',
    long_description=read("README.md"),
    long_description_content_type='text/markdown',
    packages=find_packages(exclude=['tests']),
    include_package_data=True,
    zip_safe=False,
    platforms='any',
    install_requires=dependencies,
    entry_points={
        'console_scripts': [
            'etl-script = etl_script.cli:main',
        ],
    },
    classifiers=[
        # As from http://pypi.python.org/pypi?%3Aaction=list_classifiers
        # 'Development Status :: 1 - Planning',
        # 'Development Status :: 2 - Pre-Alpha',
        # 'Development Status :: 3 - Alpha',
        'Development Status :: 4 - Beta',
        # 'Development Status :: 5 - Production/Stable',
        # 'Development Status :: 6 - Mature',
        # 'Development Status :: 7 - Inactive',
        'Environment :: Console',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: BSD License',
        'Operating System :: POSIX',
        'Operating System :: MacOS',
        'Operating System :: Unix',
        'Operating System :: Microsoft :: Windows',
        'Programming Language :: Python',
        'Programming Language :: Python :: 2',
        'Programming Language :: Python :: 3',
        'Topic :: Software Development :: Libraries :: Python Modules',
    ]
)
